/***********************************************************
	brain.c

        Copyright (c) 1997, Particle

	An AI project.
	(10/02/1997,10/04/1997,10/05/1997,10/08/1997,
	10/10/1997,10/13/1997,10/21/1997)

	Will NOT compile under Borland.  Needs a 32-bit
	Protected Mode to run & compile. (it uses
	too much memory for real mode DOS)

	Compiled with DJGPP v2.   DJGPP is a FREE 32-bit
	Protected Mode Compiler for DOS, and can be downloaded
	from http://www.delorie.com

	Will also compile under UNIX.
	(to compile under UNIX, re-save the file, because DOS stores
	'\n' character as 2 characters while UNIX stores it as 1
	character.  Re-saving the file again, eliminates the problem.)

	Particle
	bsptree@geocities.com
	http://www.geocities.com/SiliconValley/Way/7650
************************************************************/

/* includes */

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<assert.h>

/***********************************************************
defines
************************************************************/

/* printing out the output as going along, takes quite a lot of
processing time, due to the fact that the resolution is implemented
using integers, and to print, those integers have to be
converted to strings... so, it's a little faster to switch off
the printing. */

	/* print or not to print resolutions of disjunctions
	as going through the input... */
#define PRINT_RESOLUTIONS 1

	/* print or not to print output to standard output in regular
	operation mode,.. (not DEBUG mode) */
#define REGULAR_OUTPUT_OPARATION 1


	/* number of disjunctions, max "lines" in txt file, and
	on either open or closed list.. the max disjunctions
	in a conjunction.*/
#define MAXDISJUNCTIONS 4096

	/* max temporary buffer size, a char string used for
	general purposes... */
#define TMPBUFFERSIZE 1024

	/* the maximum number of terms in one disjunction */
#define MAXTERMS 64

#define OPENLIST 1
#define CLOSEDLIST 0

#define NULL_TERM -1            /* end resolution, found a contradiction */

#define NOT_CHARACTER '/'

#ifndef DEBUG
	#define DEBUG 0
#endif

/********************************************************
structure definitions
*********************************************************/

 /* 10/04/1997 */
typedef struct{
	int number;
	signed int terms[MAXTERMS];
}disjunction;

	/* this is based like a simple array stack */
typedef struct{
	int number;
	disjunction terms[MAXDISJUNCTIONS];
}conjunction;

/*********************************************************
globals
**********************************************************/
	/* a buffer for general stuff... */
char buffer[TMPBUFFERSIZE];

int num_strings=1;    /* can't start with 0 */
char *strings[MAXTERMS];

conjunction open_list;
conjunction closed_list;

/************************************************************
functions...
*************************************************************/

/**************************************************************
function to add that string to a global list of currently used
strings... which are associated with an integer by a function
"find"
 (10/02/1997)
***************************************************************/
int addstring(char *string){
	strings[num_strings]=
			(char*)malloc((strlen(string)+1)*sizeof(char));
	if(!strings[num_strings])       return 0;
	strcpy(strings[num_strings],string);
	num_strings++;
	return 1;
}

/****************************************************************
function which searches for a string in a global string database.
it returns -1 if not found, or index of string if found...
note: indexes start with 1, not 0.
 (10/02/1997,10/04/1997)
*****************************************************************/
int find(char *string){
	int i;
		/* has to start from 1, not 0 */
	for(i=1;i<num_strings;i++)
		if(!strcmp(strings[i],string))
			return i;
	return -1;
}


/*************************************************************
pushes a disjunction onto a specified list...
 (10/02/1997,10/04/1997,10/13/1997)
**************************************************************/
int pushd(conjunction *where,disjunction *what){
	assert(where->number < MAXDISJUNCTIONS);
	where->terms[++where->number] = *what;
	return 1;
}

/**************************************************************
 pops a disjunction from a specified list
 (10/02/1997,10/04/1997)
***************************************************************/
disjunction popd(conjunction *where){
     #if 0    /* the code is smart enough not to let it happen */
	if(where->number == -1){
		/* big error here.. */
		fprintf(stderr,"Poping an empty stack.\n");
		exit(1);
	}
     #endif
	return where->terms[where->number--];
}

/*************************************************************
 pushes a term (int) onto a specified list...
 (10/04/1997,10/13/1997)
**************************************************************/
int pusht(disjunction *disjunct,signed int t){
	assert(disjunct->number < MAXTERMS);
	disjunct->terms[++disjunct->number] = t;
	return 1;
}

/*************************************************************
 pops a term (int) from a specified list
 (10/04/1997)
**************************************************************/
signed int popt(disjunction *where){
	return where->terms[where->number--];
}

/*************************************************************
 just peeks at a disjunction of a specified list
 (10/02/1997)
**************************************************************/
disjunction peek(conjunction *where,int index){
	return where->terms[index];
}

/*************************************************************
	prints one disjunction
	(10/10/1997)
**************************************************************/
void print_disjunction(disjunction *tmp,FILE *where)
{
	int j;
	for(j=0;j<(tmp->number+1);j++)
		if(tmp->terms[j]){
			if(tmp->terms[j] < 0)
				fprintf(where,"%c ",NOT_CHARACTER);
			fprintf(where,"%s ",strings[abs(tmp->terms[j])]);
		}
	fputc('\n',where);
}

/*************************************************************
 prints the specified list to the specified output file...
 (10/02/1997,10/04/1997)
**************************************************************/
void print_list(conjunction *what,FILE *where){
	int i,j;
	disjunction tmp;
	for(i=0;i<(what->number+1);i++){
		tmp = what->terms[i];
		for(j=0;j<(tmp.number+1);j++)
					if(tmp.terms[j]){
			if(tmp.terms[j] < 0)
				fprintf(where,"%c ",NOT_CHARACTER);
			fprintf(where,"%s ",strings[abs(tmp.terms[j])]);
					}
				fputc('\n',where);
	}
}

/**************************************************************
 given a disjunction string, it will parse the string, and put
 them onto specified list.
 (10/02/1997,10/04/1997)
***************************************************************/
void parse_term(char *string,int list){
	int i,negation;
	char *p;
	char tmpbuffer[TMPBUFFERSIZE];

	disjunction dis;
	dis.number=-1;

	while(*string){
		negation = 0;
		p = tmpbuffer;
		while(*string == ' ') string++;

		if(*string == NOT_CHARACTER){
			negation = 1;
			string++;
		}
		/* skip whitespace */
			while(*string == ' ') string++;

		/* do a copy untill whitespace or null*/
			while(*string && *string != ' ')
			*(p++) = *(string++);
                /* make it a null terminated string */
                *p = '\0';
                        /* look for string in strings database, find
                        the associated integer */
		if((i=find(tmpbuffer)) != -1){       /* found */
			if(negation) i=-i;
				pusht(&dis,i);
		}

		else			/* string not found, adding new one */
                if(addstring(tmpbuffer)){ /* after adding, search for it... */
                        if((i=find(tmpbuffer)) != -1){       /* found */
                                if(negation) i=-i;
                                pusht(&dis,i);
                        }
                }        
			else return;  /* some error occured */
        }
		if(dis.number != -1)
			if(list == CLOSEDLIST)
					pushd(&closed_list,&dis);
			else	pushd(&open_list,&dis);
}


/******************************************************************
Given a file pointer, it will load the data from the file into
the global variables... (database).
(10/02/1997,10/21/1997) 
*******************************************************************/
int load_file(FILE *input)
{
        /* 0 stands for closed list, 1 stands for open list */
        int currentlist=OPENLIST;

                /* were using an integer stack, so, intitialize it
                to empty */
        open_list.number = -1;
	closed_list.number = -1;

                /* go through the file */
        while(fgets(buffer,TMPBUFFERSIZE,input)){
                /* get rid of new-line character at end */

			buffer[strlen(buffer)-1] = '\0';

                #if 0    /* change file format */
                        /* switch between lists */
                if(!strcmp(buffer,"closed_list"))
                        currentlist = CLOSEDLIST;
                else if(!strcmp(buffer,"open_list"))
                        currentlist = OPENLIST;
                else if(!strcmp(buffer,"END"))
                        break;
                else 
                #endif
                        /* do parse on disjunction and put it into
                        the specified list...*/
		parse_term(buffer,currentlist);
	}
        return 1;       /* return 1 for success. */
}

/*******************************************************
        looks for a number in an array
        (10/08/1997)
********************************************************/
int number_find(int *array,int n,int p){
        int i;
        for(i=0;i<n;i++)
                if(p == array[i])
                        return 1;
        return 0;
}

/******************************************************************
function to resolve 2 disjunctions, and to store the resolvent
as a third disjunction.  return 0 is no resoving has been made,
return 1 if resolved some, and return NULL_TERM if completely
resolved to a null term.  NULL_TERM is basically -1.
 (10/05/1997,10/08/1997,10/10/1997,10/13/1997)
*******************************************************************/

#if 1
int resolved(disjunction *s1,disjunction *s2,disjunction *d)
{
        int i,j,k,n;
        signed int t,all[MAXTERMS+MAXTERMS+2];
        d->number = -1;
                /* make sure the mem is 0 */
        memset(all,0,sizeof(all));

        /* insert all numbers into the array, eliminate duplicates */
        for(i=j=0,k=s1->number+1;i<k;i++){
                if(!number_find(all,j,t=s1->terms[i]))
                        all[j++] = t;
        }
        for(i=0,k=s2->number+1;i<k;i++){
		if(!number_find(all,j,t=s2->terms[i]))
			all[j++] = t;
        }

        n = j;
        k = n - 1;

        for(i=0;i<k;i++){
                while(i<k && (all[i] == 0)) i++;
                for(j=i+1;j<n;j++){
                        while(j<n && (all[j] == 0)) j++;
                        /* check to see if they're negatives of each other */
                        if(!((signed int)all[i] + (signed int)all[j])){
                                all[j] = all[i] = 0;
                                break;  /* break out of this inner loop */
                        }
                }
        }
        
        for(i=0;i<n;i++)       /* only insert if not resolved */
		if(all[i] != 0)      pusht(d,all[i]);

        if(d->number == -1)
             return NULL_TERM;
        else if(d->number == k)
             return 0; /* nothing resolved */
        else return 1;          /* something resolved */
}
#endif
/************************************************************
        function to find a disjunction inside a conjunction
        to be used to eliminate duplicates, (can get quite slow)
        (10/08/1997,10/10/1997,10/13/1997)
*************************************************************/
int disjunction_find(conjunction *c,disjunction *d)
{
        disjunction tmp2,tmp3,tmp = *d;
        int i,j;
                /* negate terms */
        j = tmp.number+1;
	for(i=0;i<j;i++)
		tmp.terms[i] = -tmp.terms[i];
		/* see if it resolves with anything */
	j = c->number+1;
	for(i=0;i<j;i++){
		tmp2 = c->terms[i];
			/* if a complete resolve, than found duplicate */
		if(resolved(&tmp,&tmp2,&tmp3) == NULL_TERM)
			return 1;
	}
	return 0;
}

/******************************************************************
 resolve_all function... to resolve it all...
 (10/05/1997,10/08/1997,10/10/1997,10/13/1997)
*******************************************************************/
int resolve_all()
{
disjunction dtmp,dtmp2,dtmp3;
int i,r,j;
	while(open_list.number != -1){

#if REGULAR_OUTPUT_OPARATION
	/* print lists */
	fprintf(stdout,"************************\n");
	fprintf(stdout,"closed list: \n");
	print_list(&closed_list,stdout);
	fprintf(stdout,"\nopen list: \n");
	print_list(&open_list,stdout);
	fprintf(stdout,"------------------------\n");
#endif

		dtmp = popd(&open_list);
		for(i=0;i<(closed_list.number+1);i++){
		      dtmp2 = closed_list.terms[i]; //peek(&closed_list,i);

		      r=resolved(&dtmp,&dtmp2,&dtmp3);
		      if(r > 0){
			#if PRINT_RESOLUTIONS
				fprintf(stdout,"resolving: ");
				print_disjunction(&dtmp,stdout);
				fprintf(stdout,"with: ");
				print_disjunction(&dtmp2,stdout);
				fprintf(stdout,"to get: ");
				print_disjunction(&dtmp3,stdout);
				fflush(stdout);
			#endif
			/* only insert unique ones, no duplicates */
			if(!disjunction_find(&open_list,&dtmp3))
			   if(!disjunction_find(&closed_list,&dtmp3))
				pushd(&open_list,&dtmp3);
		      }else if(r == NULL_TERM){
			#if PRINT_RESOLUTIONS
				fprintf(stdout,"resolving: ");
				print_disjunction(&dtmp,stdout);
				fprintf(stdout,"with: ");
				print_disjunction(&dtmp2,stdout);
				fprintf(stdout,"to get: NULL_TERM\n");
				fflush(stdout);
			#endif
			return 1;     /* resolved */
		      }
		}
			/* only insert unique ones, no duplicates */
		if(!disjunction_find(&closed_list,&dtmp))
			if(!disjunction_find(&open_list,&dtmp))
				pushd(&closed_list,&dtmp);
	}
	return 0;       /* not resolved, went through everything */
}

/******************************************************************
 main
 (10/02/1997,10/08/1997,10/10/1997)
*******************************************************************/
int main(){
	FILE *infile;
	fputs("Resolution Prover\n",stderr);
	fputs("Type in file name of file to resolve: ",stderr);

	if(!fgets(buffer,TMPBUFFERSIZE,stdin)){
		fprintf(stderr,"file name can't be a NULL string.\n");
		return 1;
		}

	/* get rid of new-line character at end */
	buffer[strlen(buffer)-1] = '\0';

		/* open file */
	if((infile = fopen(buffer,"rt")) == NULL){
		fprintf(stderr,"can't open file: %s\n",buffer);
		return 2;
		}

		/* get info out of file */
	if(!load_file(infile)){
		fprintf(stderr,"can't load file: %s\n",buffer);
		fprintf(stderr,"likely an incompatible file format.\n");
                return 3;
		}
        fclose(infile);   /* close file */

        /* print lists */
        fprintf(stdout,">>> STARTING <<<");
        fprintf(stdout,"\nclosed list: \n");
        print_list(&closed_list,stdout);

        fprintf(stdout,"\nopen list: \n");
        print_list(&open_list,stdout);

        if(resolve_all())
                fprintf(stdout,"\nContradiction Found!\n");
        else    fprintf(stdout,"\nNo Contradiction, Not Enough Data!\n");
        fflush(stdout);

        return 0;
}
